/*
 *  (c) 2005,2006 Frank A. Zdarsky, disco | Distributed Computer Systems Lab
 *                                  University of Kaiserslautern, Germany
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.math.functions;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.RenderingHints;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.JComponent;

/**
 * Simple and stupid class for painting objects of classes implementing the
 * <code>Function</code> or <code>PWLFunction</code> interfaces.
 * 
 * @author Frank A. Zdarsky
 */
public class FunctionView extends JComponent {
	private static final long serialVersionUID = 1L;
	
	private BufferedImage bi   = null;
	private Graphics2D    big2 = null;

	private List<Function> functions = new ArrayList<Function>();		
	private Map<Function, Color> color_map  = new HashMap<Function, Color>();

	private int width        = 0;
	private int height       = 0;
	private int border       = 40;
	private int tick_spacing = 50;
		
	private double x_scale  = 1.0;
	private double y_scale  = 1.0;
	
	private String x_label  = "";
	private String y_label  = "";
	
	private int xmark = -1;
	
	/**
	 * Adds <code>f</code> to the list of displayed functions. The function is
	 * painted in black.
	 * @param f the function to paint
	 */
	public void addFunction(Function f) {
		functions.add(f);
		color_map.put(f, Color.black);
	}

	/**
	 * Adds <code>f</code> to the list of displayed functions. The function is
	 * painted in the specified color <code>c</code>.
	 * @param f the function to paint
	 * @param c the color to paint function in
	 */
	public void addFunction(Function f, Color c) {
		functions.add(f);
		color_map.put(f, c);
	}

	/**
	 * Removes <code>f</code> from the list of displayed functions.
	 * @param f the function to remove
	 */
	public void removeFunction(Function f) {
		functions.remove(f);
		color_map.remove(f);
	}
	
	/**
	 * Sets the scale factors for the x- and y-axis.
	 * @param x_scale
	 * @param y_scale
	 */
	public void setAxisScales(double x_scale, double y_scale) {
		this.x_scale = x_scale;
		this.y_scale = y_scale;
	}
	
	/**
	 * Sets the labels for the x- and y-axis.
	 * @param x_label
	 * @param y_label
	 */
	public void setAxisLabels(String x_label, String y_label) {
		this.x_label = x_label;
		this.y_label = y_label;
	}
	
	public String getXAxisLabel() {
		return x_label;
	}
	
	public String getYAxisLabel() {
		return y_label;
	}
	
	public int getBorderWidth() {
		return border;
	}
	
	public void setXMark(int pos) {
		this.xmark = pos;
	}
	
	public void update() {
		updateView(big2);
		repaint();
	}
	
	public void updateSize() {
        Insets insets = getInsets();
    	width  = getWidth() - insets.left - insets.right;
    	height = getHeight() - insets.top - insets.bottom;
        
        bi   = (BufferedImage)createImage(width, height);
        big2 = bi.createGraphics();
        
        updateView(big2);
	}
	
	public void updateView(Graphics2D g2d) {
		g2d.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
		
		g2d.setColor(Color.white);
		g2d.fillRect(0, 0, width-1, height-1);
		g2d.setColor(Color.black);
		drawCoordinateSystem(g2d, border, height - 1 - border, width-2*border, height-2*border);
		drawAxisLabels(g2d, border, height - 1 - border, width-2*border, height-2*border);
		drawTicks(g2d, border, height - 1 - border, width-2*border, height-2*border);
		
		for (int i = 0; i < functions.size(); i++) {
			Function f = (Function) functions.get(i);
			if (f == null) {
				continue;
			}
			
			g2d.setColor((Color) color_map.get(f));

			int x_exit  = 0;
			int x_start = Math.max(0, (int) (f.getMinX()/x_scale));
			int x_end   = Math.min(width-2*border-1, (int) (f.getMaxX()/x_scale));
			for (int x = x_start; x <= x_end; x++) {
				int y = (int) (f.f(x*x_scale)/y_scale);
				int px = border + x;
				int py = height - border - 1 - y;
				g2d.drawLine(px, py, px, py);
				if (py >= 0) {
					x_exit = px;
				}
			}
			
			if (PWLFunction.class.isInstance(f)) {
				PWLFunction c = (PWLFunction) f;
				for (int j = 0; j < c.getSegmentCount(); j++) {
					int x = (int) (c.getIPX(j)/x_scale);
					int y = (int) (c.getIPY(j)/y_scale);
					drawMarker(g2d, border + x, height-border-1 - y, c.isLeftopen(j));
				}

				int x_label_pos = Math.min(x_exit - 100, width - border - 100);
				int y_label_pos = height - border - 1;
				y_label_pos -= (int) (f.f((x_label_pos-border)*x_scale)/y_scale);
				
				AffineTransform saveXform = g2d.getTransform();
				g2d.transform(AffineTransform.getTranslateInstance(x_label_pos-5, y_label_pos-5));
				g2d.transform(AffineTransform.getRotateInstance(
						-Math.atan(c.getGradientAt((x_label_pos-border)*x_scale)) ));
				g2d.drawString(c.getAlias(), 0, 0);
				g2d.setTransform(saveXform);
			}
		}
	}
	
	private void drawCoordinateSystem(Graphics2D g2d, int x0, int y0, int width, int height) {
		// draw x-axis
		g2d.drawLine(x0, y0, x0 + width - 1, y0);
		g2d.drawLine(x0 + width - 1, y0, x0 + width - 1 - 4, y0 - 4);
		g2d.drawLine(x0 + width - 1, y0, x0 + width - 1 - 4, y0 + 4);
		// draw y-axis
		g2d.drawLine(x0, y0, x0, y0 - height + 1);
		g2d.drawLine(x0, y0 - height + 1, x0 - 4, y0 - height + 1 + 4);
		g2d.drawLine(x0, y0 - height + 1, x0 + 4, y0 - height + 1 + 4);
	}
	
	private void drawAxisLabels(Graphics2D g2d, int x0, int y0, int width, int height) {
		if (x_label != null) {
			g2d.drawString(x_label, x0 + width - 30, y0 + 15);			
		}
		if (y_label != null) {
			g2d.drawString(y_label, x0 - 30, y0 - height + 20);
		}
	}
	
	private void drawTicks(Graphics2D g2d, int x0, int y0, int width, int height) {
		int x_start = 0;
		int x_ticks = width / tick_spacing;
		for (int i = 0; i < x_ticks; i++) {
			int tick_height = (i % 5 == 0) ? 8 : 4;
			g2d.drawLine(x0 + x_start + i*tick_spacing, y0,
					x0 + x_start + i*tick_spacing, y0 + tick_height);
			if (i % 5 == 0) {
				String tick_label = "" + (x_scale*i*tick_spacing);
				g2d.drawString(tick_label, x0 + x_start + i*tick_spacing - 10, y0 + 20);				
			}
		}
			
		int y_start = 0;
		int y_ticks = height / tick_spacing;
		for (int i = 0; i < y_ticks; i++) {
			int tick_height = (i % 5 == 0) ? 8 : 4;
			g2d.drawLine(x0, y0 - y_start - i*tick_spacing,
					x0 - tick_height, y0 - y_start - i*tick_spacing);
			if (i % 5 == 0) {
				String tick_label = "" + (y_scale*i*tick_spacing);
				g2d.drawString(tick_label, x0 - 30, y0 - y_start - i*tick_spacing + 5);				
			}
		}
	}
	
	private void drawMarker(Graphics2D g2d, int x, int y, boolean open) {
		if (open) {
			g2d.drawOval(x-3, y-3, 6, 6);
		} else {
			g2d.drawLine(x-3, y-3, x+3, y+3);
			g2d.drawLine(x-3, y+3, x+3, y-3);
		}
	}
	
	public void paintComponent(Graphics g) {
        super.paintComponent(g);
        
        Insets insets = getInsets();
    	int current_width  = getWidth() - insets.left - insets.right;
    	int current_height = getHeight() - insets.top - insets.bottom;
        
        if (bi == null) {
        	width  = current_width;
        	height = current_height;
            bi   = (BufferedImage)createImage(width, height);
            big2 = bi.createGraphics();
            updateView(big2);
        }
        
        if (width != current_width || height != current_height) {
        	updateSize();
        }

        ((Graphics2D) g).drawImage(bi, insets.left, insets.top, this);
        if (xmark >= 0) {
    		g.setColor(Color.lightGray);
    		g.drawLine(border + xmark, border, border + xmark, height-border-1);
        }
	}
}
