/*
 *  (c) 2005,2006 Frank A. Zdarsky, disco | Distributed Computer Systems Lab
 *                                  University of Kaiserslautern, Germany
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.nc;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.jfree.data.general.SeriesException;
import org.jfree.data.xy.XYDataItem;
import org.jfree.data.xy.XYSeries;
import unikl.disco.math.functions.PWLFunction;

/**
 * Class representing a piecewise linear curve, defined on [0,inf).<br/>
 * The curve is stored as an array of <code>LinearSegment</code>
 * objects. Each of these objects defines a linear piece of the
 * curve from one inflection point up to, but not including, the
 * next. It is possible to define discontinuities by defining two
 * subsequent <code>LinearSegment</code> instances which start at
 * the same inflection point. In this case, the second segment
 * needs to have <code>leftopen</code> set to <code>true</code> to
 * indicate that the inflection point is excluded from the second
 * segment.<br/>
 * All arithmetic operations on a curve return a new instance of class
 * <code>Curve</code>.<br/>
 *
 * @author Frank A. Zdarsky
 */
public class Curve extends BlankXYSeries implements PWLFunction {
	private static final double EPSILON = 1e-6;

	private static final int OPERATOR_ADD = 0;
	private static final int OPERATOR_SUB = 1;
	private static final int OPERATOR_MIN = 2;
	private static final int OPERATOR_MAX = 3;

	private static final boolean PERFORM_CHECKS = true; // Set to false to increase performance.

	public static final Curve NULL_CURVE = Curve.createHorizontal(0.0);
	public static final Curve ZERO_DELAY_BURST_CURVE = Curve.createBurstDelay(0.0);

	public static boolean create_labels = false;

	private LinearSegment[] segments;
	private List<Curve> token_buckets;
	private List<Curve> rate_latencies;

	private String label;
	private String alias;

	private static long instance_id = 0;

	/**
	 * Creates a horizontal curve.
	 * @param y the y-intercept of the curve
	 * @return a <code>Curve</code> instance
	 */
	public static Curve createHorizontal(double y) {
		Curve c = new Curve(1);
		c.segments[0].y0 = y;
		return c;
	}

	/**
	 * Creates a burst delay curve.
	 * @param t the delay, which must be >= 0.0
	 * @return a <code>Curve</code> instance
	 */
	public static Curve createBurstDelay(double t) {
		Curve c = new Curve(2);
		c.segments[1].x0       = Math.max(0.0, t);
		c.segments[1].y0       = Double.POSITIVE_INFINITY;
		c.segments[1].grad     = 0.0;
		c.segments[1].leftopen = true;
		if (Curve.create_labels) {
			c.label = "\u03B4" + "(" + t + ")";
		}
		return c;
	}

	/**
	 * Creates a new token bucket curve.
	 * @param b the burstiness
	 * @param r the rate
	 * @return a <code>Curve</code> instance
	 */
	public static Curve createTokenBucket(double b, double r) {
		Curve c = new Curve(2);
		c.segments[1].x0       = 0.0;
		c.segments[1].y0       = b;
		c.segments[1].grad     = r;
		c.segments[1].leftopen = true;
		if (Curve.create_labels) {
			c.label = "\u03B3" + "(" + b + "," + r + ")";
		}
		return c;
	}

	/**
	 * Creates a new rate latency curve.
	 * @param t the latency
	 * @param r the rate
	 * @return a <code>Curve</code> instance
	 */
	public static Curve createRateLatency(double t, double r) {
		Curve c;
		if (t == 0.0) {
			c = new Curve(1);
			c.segments[0].grad = r;
		} else {
			c = new Curve(2);
			c.segments[1].x0   = t;
			c.segments[1].grad = r;
		}
		if (Curve.create_labels) {
			c.label = "\u03B2" + "(" + t + "," + r + ")";
		}
		return c;
	}

	/**
	 * Creates a new curve from a list of token bucket curves.
	 * @param token_buckets a list of token bucket curves
	 * @return a <code>Curve</code> instance
	 */
	public static Curve createFromTokenBuckets(List token_buckets) {
		Curve c = Curve.ZERO_DELAY_BURST_CURVE;
		for (Iterator tb_iter = token_buckets.iterator(); tb_iter.hasNext(); ) {
			c = c.min((Curve) tb_iter.next());
		}
		return c;
	}

	/**
	 * Creates an empty <code>Curve</code> instance.
	 */
	private Curve() {
		super(new Long(++instance_id));
		segments = new LinearSegment[0];
	}

	/**
	 * Creates a <code>Curve</code> instance with <code>segment_count</code>
	 * empty <code>LinearSegment</code> instances.
	 * @param segment_count the number of segments
	 */
	private Curve(int segment_count) {
		super(new Long(++instance_id));
		segments = new LinearSegment[segment_count];
		for (int i = 0; i < segment_count; i++) {
			segments[i] = new LinearSegment();
		}
	}

	/**
	 * Creates a <code>Curve</code> instance with inflection points at
	 * the x-coordinates <code>ip_xs</code> and with slopes defined
	 * by <code>ip_ms</code>. The y-coordinates are computed automatically.
	 * @param ip_xs
	 * @param ip_ms
	 */
	public Curve(double[] ip_xs, double[] ip_ms) {
		this(ip_xs.length);
		for (int i = 0; i < segments.length; i++) {
			segments[i].x0   = ip_xs[i];
			segments[i].grad = ip_ms[i];
		}
		markDiscontinuities();
		computeYs(1);
	}

	/**
	 * Creates a <code>Curve</code> instance with inflection points at
	 * the x-coordinates <code>ip_xs</code> and with slopes defined
	 * by <code>ip_ms</code>. This constructor accepts an array of
	 * y-coordinates so that discontinuities can be created. The
	 * parameter <code>compute_y_start</code> indicates from which
	 * inflection point (starting with 0), the y-coordinate is again
	 * computed automatically. Note that all arrays need to have the
	 * same length.
	 * @param ip_xs
	 * @param ip_ms
	 */
	public Curve(double[] ip_xs, double[] ip_ys, double[] ip_ms,
			int compute_y_start) {
		this(ip_xs.length);
		for (int i = 0; i < segments.length; i++) {
			segments[i].x0   = ip_xs[i];
			segments[i].y0   = ip_ys[i];
			segments[i].grad = ip_ms[i];
		}
		markDiscontinuities();
		if (compute_y_start >= 1) {
			computeYs(compute_y_start);
		}
	}

	/**
	 * Returns a copy of this instance.
	 * @return a copy of this instance.
	 */
	public Curve copy() {
		Curve copy = new Curve(segments.length);
		for (int i = 0; i < segments.length; i++) {
			copy.segments[i] = segments[i].copy();
		}
		copy.label = this.label; // can copy directly as strings are unmutable
		copy.alias = this.alias;
		return copy;
	}

	/**
	 * Whether to auto-create labels for curves.
	 * @param create_labels
	 */
	public static void setCreateLabels(boolean create_labels) {
		Curve.create_labels = create_labels;
	}

	public String getLabel() {
		return (label != null) ? label : "";
	}

	public void setAlias(String alias) {
		this.alias = alias;
	}

	public String getAlias() {
		return (alias != null) ? alias : getLabel();
	}

	/**
	 * Returns the number of segments in this curve.
	 * @return the number of segments
	 */
	public int getSegmentCount() {
		return segments.length;
	}

	/**
	 * Returns the x-coordinate of inflection point <code>i</code>.
	 * @param i the index of the IP.
	 * @return the x-coordinate of the inflection point.
	 */
	public double getIPX(int i) {
		return segments[i].x0;
	}

	/**
	 * Returns the y-coordinate of inflection point <code>i</code>.
	 * @param i the index of the IP.
	 * @return the y-coordinate of the inflection point.
	 */
	public double getIPY(int i) {
		return segments[i].y0;
	}

	/**
	 * Returns the sustained rate (the gradient of the last segment).
	 * @return the sustained rate.
	 */
	public double getSustainedRate() {
		return segments[segments.length-1].grad;
	}

	/**
	 * Returns the highest rate (the steepest gradient of any segment).
	 * @return the highest rate.
	 */
	public double getHighestRate() {
		double max_rate = Double.NEGATIVE_INFINITY;
		for (int i = 0; i < segments.length; i++) {
			if (!isDiscontinuity(i) && segments[i].grad > max_rate) {
				max_rate = segments[i].grad;
			}
		}
		return max_rate;
	}

	/**
	 * Sets the sustained rate (the gradient of the last segment).
	 * Note: No checks are performed when setting the sustained rate.
	 * The caller has to ensure that <code>r</code> is larger (smaller) than
	 * the gradient of the 2nd-to-last segment to keep the curve convex (concave).
	 * @param r the new sustained rate
	 */
	public void setSustainedRate(double r) {
		segments[segments.length-1].grad = r;
	}

	/**
	 * Returns whether the inflection point <code>i</code> is excluded
	 * from the segment or not.
	 * @param i the index of the IP.
	 * @return <code>true</code> if the IP is excluded, <code>false</code> if not.
	 */
	public boolean isLeftopen(int i) {
		return segments[i].leftopen;
	}

	/**
	 * Returns whether the inflection point is a (real or unreal) discontinuity.
	 * @param i the index of the IP
	 * @return <code>true</code> if the IP is a discontinuity, <code>false</code> if not.
	 */
	public boolean isDiscontinuity(int i) {
		return (i+1 < segments.length && Math.abs(segments[i+1].x0 - segments[i].x0) < EPSILON);
	}

	/**
	 * Returns whether the inflection point is a real discontinuity, i.e. the y0
	 * of the leftopen segment differs from the previous one.
	 * @param i the index of the IP
	 * @return <code>true</code> if the IP is a real discontinuity, <code>false</code> if not.
	 */
	public boolean isRealDiscontinuity(int i) {
		return (isDiscontinuity(i) && Math.abs(segments[i+1].y0 - segments[i].y0) >= EPSILON);
	}

	/**
	 * Returns whether the inflection point is an unreal discontinuity, i.e. the y0
	 * of the leftopen segment is coincident with the y0 of the previous segment
	 * and therefore the unreal discontinuity may safely be removed.
	 * @param i the index of the IP
	 * @return <code>true</code> if the IP is an unreal discontinuity, <code>false</code> if not.
	 */
	public boolean isUnrealDiscontinuity(int i) {
		return (isDiscontinuity(i) && Math.abs(segments[i+1].y0 - segments[i].y0) < EPSILON);
	}

	/**
	 * Returns whether the current curve is a burst delay curve.
	 * @return <code>true</code> if the curve is burst delay, <code>false</code> otherwise
	 */
	public boolean isBurstDelay() {
		return (segments.length == 2
				&& segments[0].x0 == 0.0 && segments[0].y0 == 0.0
				&& segments[0].grad == 0.0 && !segments[0].leftopen
				&& segments[1].x0 >= 0.0 && segments[1].y0 == Double.POSITIVE_INFINITY
				&& segments[1].grad == 0.0 && segments[1].leftopen);
	}

	/**
	 * Adds a <code>LinearSegment</code> to the end of the curve.<br/>
	 * Note: It is the user's responsibility to add segments in the
	 * order of increasing x-coordinates.
	 * @param s the segment to be added.
	 */
	public void addSegment(LinearSegment s) {
		addSegment(segments.length, s);
	}

	/**
	 * Adds a <code>LinearSegment</code> at the location <code>pos</code>
	 * of the curve.<br/>
	 * Note: It is the user's responsibility to add segments in the
	 * order of increasing x-coordinates.
	 * @param pos the index into the segment array to add the new segment.
	 * @param s the segment to be added.
	 */
	public void addSegment(int pos, LinearSegment s) {
		if (pos < 0 || pos > segments.length) {
			throw new IllegalArgumentException("Index out of bounds (pos=" + pos + ")!");
		}
		LinearSegment[] old_segments = segments;
		segments = new LinearSegment[old_segments.length + 1];
		segments[pos] = s;
		if (pos > 0) {
			System.arraycopy(old_segments, 0, segments, 0, pos);
		}
		if (pos < old_segments.length) {
			System.arraycopy(old_segments, pos, segments, pos+1, old_segments.length-pos);
		}
	}

	/**
	 * Removes the segment at position <code>pos</code>.
	 * @param pos the index of the segment to be removed.
	 */
	public void removeSegment(int pos) {
		if (pos < 0 || pos >= segments.length) {
			throw new IllegalArgumentException("Index out of bounds (pos=" + pos + ")!");
		}
		LinearSegment[] old_segments = segments;
		segments = new LinearSegment[old_segments.length - 1];
		System.arraycopy(old_segments, 0, segments, 0, pos);
		System.arraycopy(old_segments, pos+1, segments, pos, old_segments.length-pos-1);
	}

	/**
	 * Marks discontinuities (subsequent segments having the same
	 * x-coordinate) by setting <code>leftopen</code> of the second segment.
	 */
	public void markDiscontinuities() {
		for (int i = 0; i < segments.length; i++) {
			segments[i].leftopen = false;
		}
		for (int i = 1; i < segments.length; i++) {
			if (segments[i-1].x0 == segments[i].x0) {
				segments[i].leftopen = true;
			}
		}
	}

	/**
	 * Computes the y-coordinates of inflection points starting with
	 * inflection point <code>start</code>.
	 * @param start the IP at which to start computing y-coordinates.
	 */
	public void computeYs(int start) {
		if (start < 1) {
			throw new IllegalArgumentException("Value of 'start' must be >= 1!");
		}
		for (int i = start; i < segments.length; i++) {
			double dx = segments[i].x0 - segments[i-1].x0;
			segments[i].y0 = segments[i-1].y0 + dx*segments[i-1].grad;
		}
	}

	/**
	 * Decomposes this curve into a list of token bucket curves and
	 * stores this list in the curve's <code>token_buckets</code> field.<br/>
	 * Note: Curve must be concave.
	 */
	private void decomposeIntoTokenBuckets() {
		if (PERFORM_CHECKS && !this.isConcave()) {
			throw new RuntimeException("Can only decompose concave functions into token buckets.");
		}

		token_buckets = new ArrayList<Curve>();
		for (int i = 0; i < segments.length; i++) {
			if (isDiscontinuity(i)) {
				continue;
			}
			double b = segments[i].y0 - (segments[i].x0-0.0)*segments[i].grad;
			double r = segments[i].grad;
			token_buckets.add(Curve.createTokenBucket(b, r));
		}
	}

	/**
	 * Returns the number of token buckets the curve can be decomposed into.
	 * @return the number of token buckets
	 */
	public int getTBComponentCount() {
		if (token_buckets == null) {
			decomposeIntoTokenBuckets();
		}
		return token_buckets.size();
	}

	/**
	 * Returns a list of token bucket curves that this curve can be
	 * decomposed into.
	 * @return the list of token buckets
	 */
	public final List getTBComponents() {
		if (token_buckets == null) {
			decomposeIntoTokenBuckets();
		}
		return token_buckets;
	}

	/**
	 * Returns the <code>i</code>th token bucket curve that this curve can be
	 * decomposed into.
	 * @param i the number of the token bucket
	 * @return the token bucket
	 */
	public Curve getTBComponent(int i) {
		if (token_buckets == null) {
			decomposeIntoTokenBuckets();
		}
		return (Curve) token_buckets.get(i);
	}

	/**
	 * Returns the burstiness of this token bucket curve.<br/>
	 * Note: For performance reasons there are no sanity checks! Only
	 *       call this method on a valid token bucket curve!
	 * @return the burstiness
	 */
	public double getTBBurst() {
		return segments[1].y0;
	}

	/**
	 * Sets the burstiness of this token bucket curve.<br/>
	 * Note: For performance reasons there are no sanity checks! Only
	 *       call this method on a valid token bucket curve!
	 * @param b the burstiness
	 */
	public void setTBBurst(double b) {
		segments[1].y0 = b;
	}

	/**
	 * Makes the token bucket curve right-continuous.<br/>
	 * Note: For performance reasons there are no sanity checks! Only
	 *       call this method on a valid token bucket curve!
	 */
	public void makeTBRightContinuous() {
		segments[0].y0 = segments[1].y0;
	}

	/**
	 * Undoes the makeTBRightContinuous() operation.<br/>
	 * Note: For performance reasons there are no sanity checks! Only
	 *       call this method on a valid token bucket curve!
	 */
	public void undoMakeTBRightContinuous() {
		segments[0].y0 = 0.0;
	}

	/**
	 * Decomposes this curve into a list of rate latency curves and
	 * stores this list in the curve's <code>rate_latencies</code> field.<br/>
	 * Note: Curve must be convex.
	 */
	private void decomposeIntoRateLatencies() {
		if (PERFORM_CHECKS && !this.isConvex()) {
			throw new RuntimeException("Can only decompose convex functions into rate latency curves.");
		}

		rate_latencies = new ArrayList<Curve>();
		for (int i = 0; i < segments.length; i++) {
			if (segments[i].y0 == 0.0 && segments[i].grad == 0.0) {
				continue;
			}
			double l = segments[i].x0 - (segments[i].y0*segments[i].grad);
			double r = segments[i].grad;
			rate_latencies.add(Curve.createRateLatency(l, r));
		}
	}

	/**
	 * Returns the number of rate latency curves the curve can be decomposed into.
	 * @return the number of rate latency curves
	 */
	public int getRLComponentCount() {
		if (rate_latencies == null) {
			decomposeIntoRateLatencies();
		}
		return rate_latencies.size();
	}

	/**
	 * Returns a list of rate latency curves that this curve can be
	 * decomposed into.
	 * @return the list of rate latency curves
	 */
	public final List getRLComponents() {
		if (rate_latencies == null) {
			decomposeIntoRateLatencies();
		}
		return rate_latencies;
	}

	/**
	 * Returns the <code>i</code>th rate latency curve that this curve can be
	 * decomposed into.
	 * @param i the number of the rate latency curve
	 * @return the rate latency curve
	 */
	public Curve getRLComponent(int i) {
		if (rate_latencies == null) {
			decomposeIntoRateLatencies();
		}
		return (Curve) rate_latencies.get(i);
	}

	/**
	 * Removes unnecessary segments.
	 */
	public void beautify() {
		int i = 0;
		while(i < segments.length-1) {
			// remove unreal discontinuity
			if (isUnrealDiscontinuity(i)) {
				segments[i+1].leftopen = segments[i].leftopen;
				removeSegment(i);
				continue;
			}
			i++;
		}

		i = 0;
		while(i < segments.length-1) {
			// join colinear segments
			if (Math.abs(segments[i+1].grad - segments[i].grad) < EPSILON
					&& Math.abs(segments[i+1].y0 - (segments[i].y0
							+ (segments[i+1].x0-segments[i].x0)*segments[i].grad)) < EPSILON) {
				removeSegment(i+1);
				if (i+1 < segments.length && !segments[i+1].leftopen) {
					segments[i].grad = (segments[i+1].y0 - segments[i].y0)
										/ (segments[i+1].x0 - segments[i].x0);
				}
				continue;
			}
			i++;
		}

		for (i = 0; i < segments.length-1; i++) {
			if (segments[i].x0 == segments[i+1].x0) {
				segments[i].grad = 0.0;
			}
		}
	}

	/**
	 * Returns the number of the segment that defines the function
	 * value at x-coordinate <code>x</code>. The
	 * number of the segment is usually the same as the one
	 * returned by <code>getSegmentLimitRight(x)</code>, except for
	 * if a segment starts at <code>x</code> and is left-open.
	 * In this case the function returns the previous segment,
	 * rather than the current segment, as the previous segment
	 * defines <code>x</code>.
	 * @param x the x-coordinate
	 * @return the index of the segment into the array.
	 */
	public int getSegmentDefining(double x) {
		for (int i = segments.length - 1; i >= 0; i--) {
			if (segments[i].leftopen) {
				if (segments[i].x0 < x) {
					return i;
				}
			} else {
				if (segments[i].x0 <= x) {
					return i;
				}
			}
		}
		return -1;
	}
	
	public LinearSegment getSegment(int i) {
		if(i>=0) {
			return segments[i];
		} else {
			return segments[segments.length + i];
		}
	}

	/**
	 * Returns the function value at x-coordinate <code>x</code>, if
	 * <code>x>=0</code>, and <code>NaN</code> if not.
	 * @param x the x-coordinate
	 * @return the function value
	 */
	public double f(double x) {
		int i = getSegmentDefining(x);
		if (i < 0) {
			return Double.NaN;
		}
		return (x-segments[i].x0) * segments[i].grad + segments[i].y0;
	}

	/**
	 * Returns the curve's gradient at x-coordinate <code>x</code>, if
	 * <code>x>=0</code>, and <code>NaN</code> if not. Note that the
	 * gradient returned at discontinuities is <code>0.0</code>, the
	 * gradient returned at an inflection point is the gradient of the
	 * linear segment right of the inflection point, but only if the
	 * segment is not left-open.
	 * @param x the x-coordinate
	 * @return the gradient
	 */
	public double getGradientAt(double x) {
		int i = getSegmentDefining(x);
		if (i < 0) {
			return Double.NaN;
		}
		return segments[i].grad;
	}

	/**
	 * Returns the number of the segment that defines the value
	 * of the function when computing the limit to the right
	 * of the function at x-coordinate <code>x</code>. The
	 * number of the segment is usually the same as the one
	 * returned by <code>getSegmentDefining(x)</code>, except for
	 * if a segment starts at <code>x</code> and is left-open.
	 * In this case the function returns the current segment,
	 * rather than the previous segment.
	 * @param x the x-coordinate
	 * @return the index of the segment into the array.
	 */
	public int getSegmentLimitRight(double x) {
		for (int i = segments.length - 1; i >= 0; i--) {
			if (segments[i].x0 <= x) {
				return i;
			}
		}
		return -1;
	}

	/**
	 * Returns the limit to the right of the function value at
	 * x-coordinate <code>x</code>, if <code>x>=0</code>, and
	 * <code>NaN</code> if not.
	 * @param x the x-coordinate
	 * @return the function value
	 */
	public double fLimitRight(double x) {
		int i = getSegmentLimitRight(x);
		if (i < 0) {
			return Double.NaN;
		}
		return (x-segments[i].x0) * segments[i].grad + segments[i].y0;
	}

	/**
	 * Returns the gradient to the right of the function value at
	 * x-coordinate <code>x</code>, if <code>x>=0</code>, and
	 * <code>NaN</code> if not.
	 * @param x the x-coordinate
	 * @return the function value
	 */
	public double getGradientLimitRight(double x) {
		int i = getSegmentLimitRight(x);
		if (i < 0) {
			return Double.NaN;
		}
		return segments[i].grad;
	}

	/**
	 * Returns the first segment at which the function reaches the
	 * value <code>y</code>. It returns -1 if the curve never
	 * reaches this value.
	 * @param y the y-coordinate
	 * @return the segment number
	 */
	public int getSegmentFirstAtValue(double y) {
		if (segments.length == 0 || segments[0].y0 > y) {
			return -1;
		}
		for (int i = 0; i < segments.length; i++) {
			if (i < segments.length-1) {
				if (segments[i+1].y0 >= y) {
					return i;
				}
			} else {
				if (segments[i].grad > 0.0) {
					return i;
				}
			}
		}
		return -1;
	}

	/**
	 * Returns the smallest x value at which the function value is
	 * equal to <code>y</code>.
	 * @param y the y-coordinate
	 * @return the smallest x value
	 */
	public double f_inv(double y) {
		return f_inv(y, false);
	}

	/**
	 * Returns the x value at which the function value is
	 * equal to <code>y</code>. If <code>rightmost</code> is
	 * <code>true</code>, returns the rightmost x-coordinate,
	 * otherwise the leftmost coordinate.
	 * @param y the y-coordinate
	 * @return the smallest x value
	 */
	public double f_inv(double y, boolean rightmost) {
		int i = getSegmentFirstAtValue(y);
		if (i < 0) {
			return Double.NaN;
		}
		if (rightmost) {
			while(i < segments.length && segments[i].grad == 0.0) {
				i++;
			}
			if (i >= segments.length) {
				return Double.POSITIVE_INFINITY;
			}
		}
		if (segments[i].grad != 0.0) {
			return segments[i].x0 + (y - segments[i].y0) / segments[i].grad;
		} else {
			return segments[i].x0;
		}
	}

	/**
	 * Returns the smallest x for which the curve is defined.
	 * @return the lower bound of the definition interval.
	 */
	public double getMinX() {
		return 0.0;
	}

	/**
	 * Returns the largest x for which the curve is defined.
	 * @return the upper bound of the definition interval.
	 */
	public double getMaxX() {
		return Double.POSITIVE_INFINITY;
	}

	public double getMaxY(double a, double b) {
		int sa = getSegmentDefining(a);
		int sb = getSegmentDefining(b);

		double ret = 0.0;
		for(int i=sa; i<=sb; ++i) {
			double end = ( i+1 < segments.length ? Math.min(getSegment(i+1).x0, b) : b );
			ret = Math.max(ret, getSegment(i).maxAfter(a, end));
		}

		return ret;
	}

	/**
	 * Tests whether the curve is wide-sense increasing.
	 * @return whether the curve is wide-sense increasing.
	 */
	public boolean isWideSenseIncreasing() {
		double y = Double.NEGATIVE_INFINITY;
		for (int i = 0; i < segments.length; i++) {
			if (segments[i].y0 < y || segments[i].grad < 0.0) {
				return false;
			}
			y = segments[i].y0;
		}
		return true;
	}

	/**
	 * Tests whether the curve is wide-sense increasing and satisfies f(0)=0.
	 * @return whether the curve is an arrival curve.
	 */
	public boolean isArrivalCurve() {
		return (isWideSenseIncreasing() && segments[0].y0 == 0.0);
	}

	/**
	 * Tests whether the curve is convex.
	 * @return whether the curve is convex.
	 */
	public boolean isConvex() {
		return isConvexIn(0.0, Double.POSITIVE_INFINITY);
	}

	/**
	 * Tests whether the curve is convex in [a,b].
	 * @param a the lower bound of the test interval.
	 * @param b the upper bound of the test interval.
	 * @return whether the curve is convex
	 */
	public boolean isConvexIn(double a, double b) {
		double last_gradient = Double.NEGATIVE_INFINITY;

		int i_start = getSegmentDefining(a);
		int i_end   = getSegmentDefining(b);
		for (int i = i_start; i <= i_end; i++) {
			if (i_start < 0) {
				break;
			}
			if (i == i_end && segments[i].x0 == b) {
				break;
			}
			double gradient;
			if (i < segments.length-1) {
				gradient = (segments[i+1].y0 - segments[i].y0) / (segments[i+1].x0 - segments[i].x0);
			} else {
				gradient = segments[i].grad;
			}
			if (gradient < last_gradient) {
				return false;
			}
			last_gradient = gradient;
		}
		return true;
	}

	/**
	 * Tests whether the curve is concave.
	 * @return whether the curve is concave.
	 */
	public boolean isConcave() {
		return isConcaveIn(0.0, Double.POSITIVE_INFINITY);
	}

	/**
	 * Tests whether the curve is concave in [a,b].
	 * @param a the lower bound of the test interval.
	 * @param b the upper bound of the test interval.
	 * @return whether the curve is concave.
	 */
	public boolean isConcaveIn(double a, double b) {
		double last_gradient = Double.POSITIVE_INFINITY;

		int i_start = getSegmentDefining(a);
		int i_end   = getSegmentDefining(b);
		for (int i = i_start; i <= i_end; i++) {
			if (i == i_end && segments[i].x0 == b) {
				break;
			}
			double gradient;
			if (i < segments.length-1) {
				gradient = (segments[i+1].y0 - segments[i].y0) / (segments[i+1].x0 - segments[i].x0);
			} else {
				gradient = segments[i].grad;
			}
			if (gradient > last_gradient) {
				return false;
			}
			last_gradient = gradient;
		}
		return true;
	}

	/**
	 * Tests whether the curve is almost concave, i.e. it is concave once its
	 * function value is larger than 0.
	 * @return whether the curve is almost concave.
	 */
	public boolean isAlmostConcave() {
		double last_gradient = Double.POSITIVE_INFINITY;

		for (int i = 0; i < segments.length; i++) {
			// skip the horizontal part at the beginning
			if (last_gradient == Double.POSITIVE_INFINITY && segments[i].grad == 0.0) {
				continue;
			}

			double gradient;
			if (i < segments.length-1) {
				gradient = (segments[i+1].y0 - segments[i].y0) / (segments[i+1].x0 - segments[i].x0);
			} else {
				gradient = segments[i].grad;
			}
			if (gradient > last_gradient) {
				return false;
			}
			last_gradient = gradient;
		}
		return true;
	}

	/**
	 * Returns a curve that is the sum of this curve and the given curve.
	 * @param other the given curve.
	 * @return the sum of curves
	 */
	public Curve add(Curve other) {
		return computeResultingCurve(this, other, OPERATOR_ADD);
	}

	/**
	 * Returns a curve that is the difference between this curve and the given curve.
	 * @param other the given curve.
	 * @return the difference between curves
	 */
	public Curve sub(Curve other) {
		return computeResultingCurve(this, other, OPERATOR_SUB);
	}

	/**
	 * Returns a curve that is the minimum of this curve and the given curve.
	 * @param other the given curve.
	 * @return the minimum of curves
	 */
	public Curve min(Curve other) {
		return computeResultingCurve(this, other, OPERATOR_MIN);
	}

	/**
	 * Returns a curve that is the maximum of this curve and the given curve.
	 * @param other the given curve.
	 * @return the maximum of curves
	 */
	public Curve max(Curve other) {
		return computeResultingCurve(this, other, OPERATOR_MAX);
	}

	/**
	 * Common helper for computing a new curve.
	 * @param c1
	 * @param c2
	 * @param operator
	 * @return
	 */
	private Curve computeResultingCurve(Curve c1, Curve c2, int operator) {
		Curve result = new Curve();

		double x = 0.0;

		double x_cross;

		boolean leftopen;

		int i1 = 0;
		int i2 = 0;
		while(i1 < c1.getSegmentCount() || i2 < c2.getSegmentCount()) {
			double x_next1 = (i1+1 < c1.getSegmentCount()) ?
								c1.segments[i1+1].x0 : Double.POSITIVE_INFINITY;
			double x_next2 = (i2+1 < c2.getSegmentCount()) ?
								c2.segments[i2+1].x0 : Double.POSITIVE_INFINITY;
			double x_next  = Math.min(x_next1, x_next2);

			leftopen = c1.segments[i1].leftopen || c2.segments[i2].leftopen;

			switch(operator) {
			case OPERATOR_ADD:
				result.addSegment(add(c1.segments[i1], c2.segments[i2], x, leftopen));
				break;
			case OPERATOR_SUB:
				result.addSegment(sub(c1.segments[i1], c2.segments[i2], x, leftopen));
				break;
			case OPERATOR_MIN:
				x_cross = c1.segments[i1].getXIntersectionWith(c2.segments[i2]);
				if (x < x_cross && x_cross < x_next) {
					result.addSegment(min(c1.segments[i1], c2.segments[i2], x, leftopen, false));
					result.addSegment(min(c1.segments[i1], c2.segments[i2], x_cross, false, true));
				} else {
					result.addSegment(min(c1.segments[i1], c2.segments[i2], x, leftopen, false));
				}
				break;
			case OPERATOR_MAX:
				x_cross = c1.segments[i1].getXIntersectionWith(c2.segments[i2]);
				if (x < x_cross && x_cross < x_next) {
					result.addSegment(max(c1.segments[i1], c2.segments[i2], x, leftopen, false));
					result.addSegment(max(c1.segments[i1], c2.segments[i2], x_cross, false, true));
				} else {
					result.addSegment(max(c1.segments[i1], c2.segments[i2], x, leftopen, false));
				}
				break;
			default:
			}

			if (x_next1 == x_next) {
				i1++;
			}
			if (x_next2 == x_next) {
				i2++;
			}
			x = x_next;
		}

		result.beautify();

		if (Curve.create_labels) {
			String label1 = c1.getAlias();
			String label2 = c2.getAlias();
			switch(operator) {
			case OPERATOR_ADD:
				result.label = "(" + ((label1.length() > 0) ? label1 : "?")
				             + "+" + ((label2.length() > 0) ? label2 : "?") + ")";
				break;
			case OPERATOR_SUB:
				result.label = "(" + ((label1.length() > 0) ? label1 : "?")
				             + "-" + ((label2.length() > 0) ? label2 : "?") + ")";
				break;
			case OPERATOR_MIN:
				result.label = "(" + ((label1.length() > 0) ? label1 : "?")
				             + "\u2227" + ((label2.length() > 0) ? label2 : "?") + ")";
				break;
			case OPERATOR_MAX:
				result.label = "(" + ((label1.length() > 0) ? label1 : "?")
				             + "\u2228" + ((label1.length() > 0) ? label1 : "?") + ")";
				break;
			default:
			}
		}

		return result;
	}

	/**
	 * Helper creating a new segment starting at x0 that is the sum of the given segments.
	 * @param s1
	 * @param s2
	 * @param x
	 * @param leftopen
	 * @return
	 */
	private LinearSegment add(LinearSegment s1, LinearSegment s2, double x, boolean leftopen) {
		LinearSegment result = new LinearSegment();
		result.x0       = x;
		result.y0       = s1.f(x) + s2.f(x);
		result.grad     = s1.grad + s2.grad;
		result.leftopen = leftopen;
		return result;
	}

	/**
	 * Helper creating a new segment starting at x0 that is the difference between the given segments.
	 * @param s1
	 * @param s2
	 * @param x
	 * @param leftopen
	 * @return
	 */
	private LinearSegment sub(LinearSegment s1, LinearSegment s2, double x, boolean leftopen) {
		LinearSegment result = new LinearSegment();
		result.x0       = x;
		result.y0       = s1.f(x) - s2.f(x);
		result.grad     = s1.grad - s2.grad;
		result.leftopen = leftopen;
		return result;
	}

	/**
	 * Helper creating a new segment starting at x0 that is the minimum of the given segments.
	 * Note: Only valid if s1 and s2 do not intersect before the next IP.
	 * @param s1
	 * @param s2
	 * @param x
	 * @param leftopen
	 * @return
	 */
	private LinearSegment min(LinearSegment s1, LinearSegment s2, double x, boolean leftopen, boolean crossed) {
		double f1_x = s1.f(x);
		double f2_x = s2.f(x);

		LinearSegment result = new LinearSegment();
		result.x0 = x;
		if (crossed || Math.abs(f1_x - f2_x) < EPSILON) {
			result.y0   = f1_x;
			result.grad = Math.min(s1.grad, s2.grad);
		} else if (f1_x < f2_x) {
			result.y0   = f1_x;
			result.grad = s1.grad;
		} else {
			result.y0   = f2_x;
			result.grad = s2.grad;
		}
		result.leftopen = leftopen;
		return result;
	}

	/**
	 * Helper creating a new segment starting at x0 that is the maximum of the given segments.
	 * Note: Only valid if s1 and s2 do not intersect before the next IP.
	 * @param s1
	 * @param s2
	 * @param x
	 * @param leftopen
	 * @return
	 */
	private LinearSegment max(LinearSegment s1, LinearSegment s2, double x, boolean leftopen, boolean crossed) {
		double f1_x = s1.f(x);
		double f2_x = s2.f(x);

		LinearSegment result = new LinearSegment();
		result.x0 = x;
		if (crossed || Math.abs(f1_x - f2_x) < EPSILON) {
			result.y0   = f1_x;
			result.grad = Math.max(s1.grad, s2.grad);
		} else if (f1_x > f2_x) {
			result.y0   = f1_x;
			result.grad = s1.grad;
		} else {
			result.y0   = f2_x;
			result.grad = s2.grad;
		}
		result.leftopen = leftopen;
		return result;
	}

	/**
	 * Returns a copy of this curve that is shifted to the right by <code>dx</code>,
	 * i.e. g(x) = f(x-dx).
	 * @param dx the offset to shift the curve.
	 * @return the shifted curve.
	 */
	public Curve shiftRight(double dx) {
		if (segments[0].y0 != 0.0) {
			throw new RuntimeException("Curve to shift right must pass through origin!");
		}
		Curve result = new Curve(segments.length + 1);
		result.segments[0] = new LinearSegment();
		result.segments[0].y0 = this.segments[0].y0; // decide what to do if pass thru origin req. dropped...
		for (int i = 0; i < segments.length; i++) {
			result.segments[i+1] = segments[i].copy();
			result.segments[i+1].x0 += dx;
		}
		result.beautify();
		if (Curve.create_labels) {
			result.label = getAlias() + "(t-" + dx + ")";
		}
		return result;
	}

	/**
	 * Returns a copy of this curve that is shifted to the left by <code>dx</code>,
	 * i.e. g(x) = f(x+dx). Note that the new curve is clipped at the y-axis so
	 * that in most cases <code>c.shiftLeftClipping(dx).shiftRight(dx) != c</code>!
	 * @param dx the offset to shift the curve.
	 * @return the shifted curve.
	 */
	public Curve shiftLeftClipping(double dx) {
		Curve result = this.copy();
		int i = result.getSegmentDefining(dx);
		if (result.segments[i].x0 < dx) {
			result.segments[i].y0 += (dx - result.segments[i].x0) * result.segments[i].grad;
			result.segments[i].x0 = dx;
			result.segments[i].leftopen = false;
		}
		if (i > 0) {
			LinearSegment[] old_segments = result.segments;
			result.segments = new LinearSegment[old_segments.length - i];
			System.arraycopy(old_segments, i, result.segments, 0, result.segments.length);
		}
		for (i = 0; i < result.segments.length; i++) {
			result.segments[i].x0 -= dx;
		}
		result.beautify();

		if (Curve.create_labels) {
			result.label = getAlias() + "(t+" + dx + ")";
		}
		return result;
	}

	/**
	 * Returns a copy of this curve with latency removed, i.e. shifted left by
	 * the latency.
	 * @return a copy of this curve without latency
	 */
	public Curve removeLatency() {
		Curve result = this.copy();

		// remove all segment(s) with y0==0.0 and grad==0.0
		while(result.segments.length > 0) {
			if (result.segments[0].y0 > 0.0 || result.segments[0].grad > 0.0) {
				break;
			}
			if (segments[0].y0 < 0.0 || segments[0].grad < 0.0) {
				throw new RuntimeException("Should have avoided neg. gradients elsewhere...");
			}
			result.removeSegment(0);
		}

		// in case that we've removed everything, the curve had infinite
		// latency, so return the NULL curve.
		if (result.segments.length == 0) {
			return Curve.NULL_CURVE;
		}

		// shift remaining segments left by latency
		double L = result.segments[0].x0;
		for (int i = 0; i < result.segments.length; i++) {
			result.segments[i].x0 -= L;
		}
		if (result.segments[0].leftopen) {
			result.addSegment(0, new LinearSegment());
		}
		result.beautify();
		if (Curve.create_labels) {
			result.label = getAlias() + "(t+" + L + ")";
		}
		return result;
	}

	/**
	 * Returns a copy of this curve bounded at the x-axis.
	 * @return the bounded curve.
	 */
	public Curve boundAtXAxis() {
		Curve result = new Curve();

		LinearSegment s;
		for (int i = 0; i < segments.length; i++) {
			if (segments[i].y0 > 0.0) {
				result.addSegment(segments[i].copy());

				if (segments[i].grad < 0.0) {
					double x_cross = segments[i].getXIntersectionWith(LinearSegment.X_AXIS);
					if (i+1 >= segments.length || x_cross < segments[i+1].x0) {
						s = new LinearSegment();
						s.x0 = x_cross;
						result.addSegment(s);
					}
				}
			} else {
				s = new LinearSegment();
				s.x0       = segments[i].x0;
				s.leftopen = segments[i].leftopen;
				result.addSegment(s);

				if (segments[i].grad > 0.0) {
					double x_cross = segments[i].getXIntersectionWith(LinearSegment.X_AXIS);
					if (i+1 >= segments.length || x_cross < segments[i+1].x0) {
						s = new LinearSegment();
						s.x0   = x_cross;
						s.grad = segments[i].grad;
						result.addSegment(s);
					}
				}
			}
		}
		result.beautify();
		if (Curve.create_labels) {
			result.label = "[" + getAlias() + "]\u207A";
		}
		return result;
	}

	/**
	 * Returns a copy of this curve shifted vertically by <code>dy</code>.
	 * @param dy the offset to shift the curve.
	 * @return the shifted curve.
	 */
	public Curve add(double dy) {
		Curve result = this.copy();
		for (int i = 0; i < segments.length; i++) {
			result.segments[i].y0 += dy;
		}
		if (Curve.create_labels) {
			result.label = "(" + getAlias() + ((dy >= 0.0) ? "+" : "") + dy + ")";
		}
		return result;
	}

	/**
	 * Returns the x-coordinate of the inflection point after which the function
	 * values are greater than zero.
	 * @return the latency of this curve.
	 */
	public double getLatency() {
		if (segments[0].y0 > 0.0) {
			return 0.0;
		}
		for (int i = 0; i < segments.length; i++) {
			double y0 = segments[i].y0;
			if (y0 < 0.0 && y0 > -EPSILON) {
				y0 = 0.0;
			}
			if (y0 > 0.0 || (y0 >= 0.0 && segments[i].grad > 0.0)) {
				return segments[i].x0;
			}
			if (y0 < 0.0 || segments[i].grad < 0.0) {
				throw new RuntimeException("Should have avoided neg. gradients elsewhere...");
			}
		}
		return Double.POSITIVE_INFINITY;
	}

	/**
	 * Returns the convolution of this curve, which must be (almost) concave, and
	 * the given curve, which must also be (almost) concave.
	 * @param other the curve to convolve with.
	 * @return the convolved curve.
	 */
	public Curve convolveAlmostConcave(Curve other) {
		if (PERFORM_CHECKS && (!this.isAlmostConcave() || !other.isAlmostConcave())) {
			throw new IllegalArgumentException("Both curves to be convolved must be almost concave!");
		}

		double L1 = this.getLatency();
		double L2 = other.getLatency();
		if (L1 == Double.POSITIVE_INFINITY) {
			return other.copy();
		}
		if (L2 == Double.POSITIVE_INFINITY) {
			return this.copy();
		}

		Curve result = (this.removeLatency()).min(other.removeLatency()).shiftRight(L1+L2);
		result.beautify();
		if (Curve.create_labels
				&& (this.label != null || this.alias != null
						|| other.label != null || other.alias != null)) {
			String label1 = (this.getAlias().length() > 0) ? this.getAlias() : "?";
			String label2 = (other.getAlias().length() > 0) ? other.getAlias() : "?";
			result.label = "(" + label1 + "\u2297" + label2 + ")";
		}
		return result;
	}

	/**
	 * Returns the convolution of this curve, which must be convex, and
	 * the given curve, which must also be convex.
	 * @param other the curve to convolve with.
	 * @return the convolved curve.
	 */
	public Curve convolveConvex(Curve other) {
		if (this.isBurstDelay() && segments[1].x0 == 0.0) {
			return other.copy();
		}
		if (other.isBurstDelay() && other.segments[1].x0 == 0.0) {
			return this.copy();
		}

		if (PERFORM_CHECKS && (!this.isConvex() || !other.isConvex())) {
			throw new IllegalArgumentException("Both curves to be convolved must be convex!");
		}

		Curve result = new Curve();

		LinearSegment s;
		s = new LinearSegment();
		s.x0 = 0.0;
		s.y0 = this.f(0.0) + other.f(0.0);
		result.addSegment(s);

		int i1 = (this.isRealDiscontinuity(0)) ? 1 : 0;
		int i2 = (other.isRealDiscontinuity(0)) ? 1 : 0;
		if (i1 > 0 || i2 > 0) {
			s = new LinearSegment();
			s.x0       = 0.0;
			s.y0       = this.fLimitRight(0.0) + other.fLimitRight(0.0);
			s.leftopen = true;
			result.addSegment(s);
		}

		while(i1 < this.getSegmentCount() || i2 < other.getSegmentCount()) {
			if (this.segments[i1].grad < other.segments[i2].grad) {
				if (i1+1 >= this.getSegmentCount()) {
					result.segments[result.segments.length-1].grad = this.segments[i1].grad;
					break;
				}

				s = new LinearSegment();
				s.x0 = result.segments[result.segments.length-1].x0
						+ (this.segments[i1+1].x0 - this.segments[i1].x0);
				s.y0 = result.segments[result.segments.length-1].y0
						+ (this.segments[i1+1].y0 - this.segments[i1].y0);
				result.segments[result.segments.length-1].grad = this.segments[i1].grad;
				result.addSegment(s);
				i1++;
			} else {
				if (i2+1 >= other.getSegmentCount()) {
					result.segments[result.segments.length-1].grad = other.segments[i2].grad;
					break;
				}

				s = new LinearSegment();
				s.x0 = result.segments[result.segments.length-1].x0
						+ (other.segments[i2+1].x0 - other.segments[i2].x0);
				s.y0 = result.segments[result.segments.length-1].y0
						+ (other.segments[i2+1].y0 - other.segments[i2].y0);
				result.segments[result.segments.length-1].grad = other.segments[i2].grad;
				result.addSegment(s);
				i2++;
			}
		}
		result.beautify();
		if (Curve.create_labels
				&& (this.label != null || this.alias != null
						|| other.label != null || other.alias != null)) {
			String label1 = (this.getAlias().length() > 0) ? this.getAlias() : "?";
			String label2 = (other.getAlias().length() > 0) ? other.getAlias() : "?";
			result.label = "(" + label1 + "\u2297" + label2 + ")";
		}
		return result;
	}

	/**
	 * Returns the deconvolution of this curve, which must be (almost) concave, and
	 * the given curve, which must convex.
	 * @param other the curve to convolve with.
	 * @return the convolved curve.
	 */
	public Curve deconvolve(Curve other) {
		if (PERFORM_CHECKS && !this.isAlmostConcave()) {
			throw new IllegalArgumentException("First curve of deconvolution must be almost concave!");
		}
		if (PERFORM_CHECKS && !other.isConvex()) {
			throw new IllegalArgumentException("Second curve of deconvolution must be convex!");
		}

		int i_this  = 0;
		int i_other = 0;

		// advance this_iterator to concave part
		while(i_this < this.getSegmentCount()) {
			if (this.segments[i_this].grad > 0.0) {
				break;
			}
			i_this++;
		}
		// advance other_iterator to first IP after the start of this's concave part
		while(i_other < other.getSegmentCount()) {
			if (other.segments[i_other].x0 >= this.segments[i_this].x0) {
				break;
			}
			i_other++;
		}

		// from here on, search both curve's IPs until grad(other)>=grad(this)
		double kj = Double.POSITIVE_INFINITY;
		while(i_this < this.getSegmentCount() || i_other < other.getSegmentCount()) {
			double x_this = (i_this < this.getSegmentCount()) ?
					this.segments[i_this].x0 : Double.POSITIVE_INFINITY;
			double x_other = (i_other < other.getSegmentCount()) ?
					other.segments[i_other].x0 : Double.POSITIVE_INFINITY;

			if (x_other <= x_this) {
				if (other.segments[i_other].grad
						>= this.getGradientLimitRight(other.segments[i_other].x0)) {
					kj = other.segments[i_other].x0;
					break;
				}
				i_other++;
			} else {
				if (other.getGradientLimitRight(this.segments[i_this].x0)
						>= this.segments[i_this].grad) {
					kj = this.segments[i_this].x0;
					break;
				}
				i_this++;
			}
		}

		if (kj < Double.POSITIVE_INFINITY) {
			Curve result = this.shiftLeftClipping(kj).add(-other.f(kj));
			result.beautify();
			if (Curve.create_labels
					&& (this.label != null || this.alias != null
							|| other.label != null || other.alias != null)) {
				String label1 = (this.getAlias().length() > 0) ? this.getAlias() : "?";
				String label2 = (other.getAlias().length() > 0) ? other.getAlias() : "?";
				result.label = "(" + label1 + "\u2298" + label2 + ")";
			}
			return result;
		} else {
			Curve c = new Curve(1);
			c.segments[0].y0 = Double.POSITIVE_INFINITY;
			return c;
		}
	}

	/**
	 * Returns a copy of this curve ensuring that f(0)=0.
	 * @return a copy with with f(0) set to 0
	 */
	public Curve makeArrivalCurve() {
		if (segments[0].y0 > 0.0) {
			Curve result = new Curve(segments.length+1);
			System.arraycopy(segments, 0, result.segments, 1, segments.length);
			result.segments[1].leftopen = true;
			result.label = "" + this.label;
			return result;
		} else {
			return this.copy();
		}
	}

	/**
	 * Computes the effetive FIFO service curve for a server with the service curve
	 * <code>beta</code> experiencing cross-traffic with arrival curve <code>alpha</code>.
	 * @param alpha arrival curve of cross-traffic
	 * @param beta server's service curve
	 * @return the FIFO service curve
	 */
	public static Curve computeFIFOServiceCurve(Curve alpha, Curve beta) {
		// double check; can be deleted later on for increased performance
		if (PERFORM_CHECKS && !alpha.isConcave()) {
			throw new IllegalArgumentException("First curve must be concave!");
		}
		if (PERFORM_CHECKS && !beta.isConvex()) {
			throw new IllegalArgumentException("Second curve must be convex!");
		}

		double last_y = alpha.fLimitRight(0.0);

		ArrayList ycoords = computeInflectionPointsY(alpha, beta);
		for(int i = 0; i < ycoords.size(); i++) {
			double ip_y = ((Double)ycoords.get(i)).doubleValue();
			if (ip_y < last_y) {
				continue;
			}

			double x_alpha = alpha.f_inv(ip_y, false);
			double x_beta  = beta.f_inv(ip_y, true);
			if (alpha.getGradientLimitRight(x_alpha) <= beta.getGradientLimitRight(x_beta)) {
				Curve beta_fifo;

				double theta = x_beta - x_alpha;
	    		beta_fifo = beta.sub(alpha.shiftRight(theta));
	    		beta_fifo = beta_fifo.min(Curve.createBurstDelay(x_beta)).boundAtXAxis();
	    		return beta_fifo;
			}
		}
		return beta.copy();
	}

	/**
	 * Returns the backlog bound of the given two curves.
	 * @param alpha the arrival curve.
	 * @param beta  the service curve.
	 * @return the value of the backlog bound.
	 */
	public static double getBacklogBound(Curve alpha, Curve beta) {
		return Math.max(0.0, getVerticalDeviation(alpha, beta));
	}

	/**
	 * Returns the vertical deviation between the given two curves.
	 * @param c1 the first curve.
	 * @param c2 the second curve.
	 * @return the value of the vertical deviation.
	 */
	public static double getVerticalDeviation(Curve c1, Curve c2) {
		if (c1.getSustainedRate() > c2.getSustainedRate()) {
			return Double.POSITIVE_INFINITY;
		}

		double result = Double.NEGATIVE_INFINITY;
		ArrayList xcoords = computeInflectionPointsX(c1, c2);
		for(int i = 0; i < xcoords.size(); i++) {
			double ip_x = ((Double)xcoords.get(i)).doubleValue();

			double backlog = c1.f(ip_x) - c2.f(ip_x);
			result = Math.max(result, backlog);
		}
		return result;
	}

	/**
	 * Returns the delay bound of the given two curves.
	 * @param alpha the arrival curve.
	 * @param beta  the service curve.
	 * @return the value of the delay bound.
	 */
	public static double getDelayBound(Curve alpha, Curve beta) {
		return Math.max(0.0, getHorizontalDeviation(alpha, beta));
	}

	/**
	 * Returns the horizontal deviation between the given two curves.
	 * @param c1 the first curve.
	 * @param c2 the second curve.
	 * @return the value of the horizontal deviation.
	 */
	public static double getHorizontalDeviation(Curve c1, Curve c2) {
		if (c1.getSustainedRate() > c2.getSustainedRate()) {
			return Double.POSITIVE_INFINITY;
		}

		double result = Double.NEGATIVE_INFINITY;
		for(int i = 0; i < c1.segments.length; i++) {
			double ip_y = c1.segments[i].y0;

			double delay = c2.f_inv(ip_y, true) - c1.f_inv(ip_y, false);
			result = Math.max(result, delay);
		}
		for(int i = 0; i < c2.segments.length; i++) {
			double ip_y = c2.segments[i].y0;

			double delay = c2.f_inv(ip_y, true) - c1.f_inv(ip_y, false);
			result = Math.max(result, delay);
		}
		return result;
	}

	/**
	 * Returns an <code>ArrayList</code> instance of those x-coordinates
	 * at which either c1 or c2 or both have an inflection point. There
	 * will be multiple occurences of an x-coordinate, if at least one
	 * curve has a discontinuity at that x-coordinate.
	 * @param c1 the first curve.
	 * @param c2 the second curve.
	 * @return an <code>ArrayList</code> of <code>Double</code> objects
	 * containing the x-coordinates of the respective inflection point.
	 */
	public static ArrayList computeInflectionPointsX(Curve c1, Curve c2) {
		ArrayList<Double> xcoords = new ArrayList<Double>();

		int i1 = 0;
		int i2 = 0;
		while(i1 < c1.getSegmentCount() || i2 < c2.getSegmentCount()) {
			double x1 = (i1 < c1.getSegmentCount()) ?
						c1.segments[i1].x0 : Double.POSITIVE_INFINITY;
			double x2 = (i2 < c2.getSegmentCount()) ?
						c2.segments[i2].x0 : Double.POSITIVE_INFINITY;
			if (x1 < x2) {
				xcoords.add(new Double(x1));
				i1++;
			} else if (x1 > x2) {
				xcoords.add(new Double(x2));
				i2++;
			} else {
				xcoords.add(new Double(x1));
				i1++;
				i2++;
			}
		}
		return xcoords;
	}

	/**
	 * Returns an <code>ArrayList</code> instance of those y-coordinates
	 * at which either c1 or c2 or both have an inflection point.
	 * @param c1 the first curve.
	 * @param c2 the second curve.
	 * @return an <code>ArrayList</code> of <code>Double</code> objects
	 * containing the x-coordinates of the respective inflection point.
	 */
	public static ArrayList computeInflectionPointsY(Curve c1, Curve c2) {
		ArrayList<Double> ycoords = new ArrayList<Double>();

		int i1 = 0;
		int i2 = 0;
		while(i1 < c1.getSegmentCount() || i2 < c2.getSegmentCount()) {
			double y1 = (i1 < c1.getSegmentCount()) ?
						c1.segments[i1].y0 : Double.POSITIVE_INFINITY;
			double y2 = (i2 < c2.getSegmentCount()) ?
						c2.segments[i2].y0 : Double.POSITIVE_INFINITY;
			if (y1 < y2) {
				ycoords.add(new Double(y1));
				i1++;
			} else if (y1 > y2) {
				ycoords.add(new Double(y2));
				i2++;
			} else {
				ycoords.add(new Double(y1));
				i1++;
				i2++;
			}
		}

		return ycoords;
	}

	/**
	 * Returns a list containing each segment's height when projected onto
	 * the y-axis.
	 * @return a list of doubles containing the segment heights
	 */
	public List<Double> getYIntervals() {
		List<Double> y_intervals = new ArrayList<Double>();
		for (int i = 1; i < segments.length; i++) {
			if (isDiscontinuity(i-1)) {
				continue;
			}
			y_intervals.add(new Double(segments[i].y0 - segments[i-1].y0));
		}
		return y_intervals;
	}

	/**
	 * Approximates arrival curve by removing all segments whose projection onto
	 * the y-axis has a height smaller than <code>dy</code>. Returns the number of
	 * removed segments.
	 * @param dy the height below which segments are removed
	 * @return the number of removed segments
	 */
	public int approximateSegments(double dy) {
		int segments_removed = 0;
		for (int i = segments.length-1; i > 0; i--) {
			if (isDiscontinuity(i-1)) {
				continue;
			}
			if (segments[i].y0 - segments[i-1].y0 <= dy) {
				approximateAtSegment(i-1);
				segments_removed++;
				break;
			}
		}
		return segments_removed;
	}

	/**
	 * Approximates a (concave) arrival curve by removing the specified segment.
	 * @param i the index of the segment to remove
	 */
	public void approximateAtSegment(int i) {
		if (PERFORM_CHECKS && !isConcave()) {
			throw new IllegalArgumentException("Curve to approximate must be concave!");
		}

		if (segments.length <= 2 || i < 1 || i >= segments.length-1) {
			return;
		}

		removeSegment(i);

		double x_cross;
		if (i == 1) {
			x_cross = 0.0;
		} else {
			x_cross = segments[i-1].getXIntersectionWith(segments[i]);
		}
		segments[i].y0      -= (segments[i].x0 - x_cross) * segments[i].grad;
		segments[i].x0       = x_cross;
		segments[i].leftopen = (i == 1);
	}

	/**
	 * Returns a string representation of this curve.
	 * @return the curve represented as a string.
	 */
	@Override
	public String toString() {
		String result = "{";
		for (int i = 0; i < segments.length; i++) {
			if (i > 0) {
				result += ",";
			}
			if (segments[i].leftopen) {
				result += "!";
			}
			result += "(" + segments[i].x0 + "," + segments[i].y0 + "),"
						+ segments[i].grad;
		}
		result += "}";
		return result;
	}

	/**
	 * Prints the string representation of this curve to stdout.
	 */
	public void dump() {
		System.out.print(this);
	}

	@Override
	public XYDataItem getDataItem(int index) { return segments[index]; }
	@Override
	public int getMaximumItemCount() { return segments.length; }
	@Override
	public java.lang.Number getX(int index) { return segments[index].x0; }
	@Override
	public java.lang.Number getY(int index) { return segments[index].y0; }

	// XYSeries interface, useful methods. Hack something that's semantically
	// close to the truth
	@Override
	public XYSeries createCopy(int start, int end) {
		XYSeries ret = new XYSeries(instance_id);
		for(int i=start; i<end; ++i) {
			ret.add(segments[i].x0, segments[i].y0);
		}

		return ret;
	}

	@Override
	public int getItemCount() { return segments.length; }

	@Override
	public List getItems() {
		return Collections.unmodifiableList(Arrays.asList(segments));
	}

	@Override
	public int indexOf(java.lang.Number x) {
		int s = getSegmentDefining(x.doubleValue());
		if(segments[s].x0 == x.doubleValue()) {
			return s;
		} else {
			return -1;
		}
	}

	//TODO: we may want to override toArray if it is used. With periodic curves,
	// it will break because it uses getItemCount for array size

	@Override
	public void update(java.lang.Number x, java.lang.Number y) {
		int i = indexOf(x);

		if(i<0) {
			throw new SeriesException("No series item at " + x);
		}

		//TODO: we don't actually update
	}
}