/*
 *  (c) 2005,2006 Frank A. Zdarsky, disco | Distributed Computer Systems Lab
 *                                  University of Kaiserslautern, Germany
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.nc;

import org.jfree.data.xy.XYDataItem;

/**
 * Class representing linear segments of a curve. A linear segments starts at
 * point (<code>x0</code>, <code>y0</code>) and continues to the right
 * with slope <code>grad</code>. If <code>leftopen</code> is
 * <code>true</code>, the point (<code>x0</code>,<code>y0</code>) is
 * excluded from the segment, otherwise is is included.
 *
 * @author Frank A. Zdarsky
 */
public class LinearSegment extends XYDataItem {
	public static final LinearSegment X_AXIS = new LinearSegment();

	/** The x-coordinate of the linear segment's starting point. */
	public double x0;

	/** The y-coordinate of the linear segment's starting point. */
	public double y0;

	/** The gradient of the linear segment */
	public double grad;

	/** Whether the point (<code>x0</code>, <code>y0</code>) is part of the segment or not. */
	public boolean leftopen;

	/**
	 * The default constructor.
	 */
	public LinearSegment() {
		super(0, 0);
	}

	/**
	 * A convenience constructor.
	 * @param x0
	 * @param y0
	 * @param grad
	 * @param leftopen
	 */
	public LinearSegment(double x0, double y0, double grad, boolean leftopen) {
		super(x0, y0);

		this.x0       = x0;
		this.y0       = y0;
		this.grad     = grad;
		this.leftopen = leftopen;
	}

	/**
	 * Returns a copy of this instance.
	 *
	 * @return a copy of this instance.
	 */
	public LinearSegment copy() {
		LinearSegment copy = new LinearSegment();
		copy.x0 = this.x0;
		copy.y0 = this.y0;
		copy.grad = this.grad;
		copy.leftopen = this.leftopen;
		return copy;
	}

	/**
	 * Returns the function value of this linear segment at the given
	 * x-coordinate. Note that there is no test whether the function is defined
	 * at this location, but simply returns the the value of the co-linear line.
	 *
	 * @param x the coorinate whose function value shall be returned
	 * @return the function value
	 */
	public double f(double x) {
		return (x - x0) * grad + y0;
	}

	/**
	 * Returns the x-coordinate at which a co-linear line through this segment
	 * intersects a co-linear line through the segment <code>other</code>.
	 *
	 * @param other the other segment
	 * @return the x-coordinate at which the segments cross
	 */
	public double getXIntersectionWith(LinearSegment other) {
		double y1 = this.y0 - this.x0 * this.grad;
		double y2 = other.y0 - other.x0 * other.grad;
		return (y2 - y1) / (this.grad - other.grad); // if lines are parallel returns NaN!
	}

	/** Maximum value inside [a,b).
	 *
	 * @param a
	 * @param b
	 * @return
	 */
	public double maxAfter(double a, double b) {
		a = Math.max(a, x0);

		return Math.max(f(a), f(b));
	}


	// Implement XYDataItem
	@Override
	public java.lang.Number getX() { return x0; }
	@Override
	public double getXValue() { return x0; }
	@Override
	public java.lang.Number getY() { return y0; }
	@Override
	public double getYValue() { return y0; }
	@Override
	public void setY(double y) { this.y0 = y; }
	@Override
	public void setY(java.lang.Number y) { this.y0 = y.doubleValue(); }
}
